#include <cstdio>

#include "instrument.hh"

using namespace std;

vector<sample> repository;

SampleMatchType TreeCompare(
  const sample &book, const deque<sampleitem> &booktree,
  const sample &,     const deque<sampleitem> &datatree)
{
    typedef deque<sampleitem>::const_iterator itemit;

    const bool dutytest = book.dutytest();
    const bool volutest = book.volutest();
    const bool wavetest = book.wavetest();
    
    itemit bookit = booktree.begin();
    itemit datait = datatree.begin();
    
    // Note: data has all "types" undefined.
    // If the book has them defined, they master the comparison.
    
    SampleMatchType match;
    
    while(bookit != booktree.end()
       && datait != datatree.end())
    {
        if(dutytest && bookit->duty != datait->duty) return false;
        if(wavetest)
        {
            const bool addmatch = bookit->wavelenadd == datait->wavelenadd;
            const bool mulmatch = bookit->wavelenmul == datait->wavelenmul;
/*
            fprintf(stderr, "[Len %u, %u%s%u %ld%s%ld]\n", match.length,
                bookit->wavelenadd,addmatch?"==":"!=", datait->wavelenadd,
                bookit->wavelenmul,mulmatch?"==":"!=", datait->wavelenmul);
*/
            if(!addmatch) { if(book.wavelentype == sample::add)return false; if(!match.wavemulmatch)return false; match.waveaddmatch = false; }
            if(!mulmatch) { if(book.wavelentype == sample::mul)return false; if(!match.waveaddmatch)return false; match.wavemulmatch = false; }
        }
        if(volutest)
        {
            const bool addmatch = bookit->voladd == datait->voladd;
            const bool mulmatch = bookit->volmul == datait->volmul;
/*
            fprintf(stderr, "[Len %u, %u%s%u %ld%s%ld]\n", match.length,
                bookit->voladd,addmatch?"==":"!=", datait->voladd,
                bookit->volmul,mulmatch?"==":"!=", datait->volmul);
*/
            if(!addmatch) { if(book.voltype == sample::add)return false; if(!match.volmulmatch)return false; match.voladdmatch = false; }
            if(!mulmatch) { if(book.voltype == sample::mul)return false; if(!match.voladdmatch)return false; match.volmulmatch = false; }
        }
        ++bookit;
        ++datait;
        ++match.length;
        #if 0
        if(match.length >= 7)
        {
            while(bookit != book.tree.end()
               && datait != data.tree.end())
            {
                ++bookit;
                ++datait;
                ++match.length;
            }
            break;
        }
        #endif
    }
    // Found a match. {wave,vol}{add,mul}match could be used
    // now to tell which kind of matching we should be doing.
    return match;
}

struct SampleMatchPairType
{
    SampleMatchType alku;
    /*SampleMatchType loppu;*/
    
    SampleMatchPairType() :  alku(false)/*, loppu(false)*/ { }
    SampleMatchPairType(const SampleMatchType &a/*,
                        const SampleMatchType &b*/) : alku(a)/*, loppu(b)*/ { }
    SampleMatchPairType(bool v) : alku(v)/*, loppu(v)*/ { }
};

// Test whether the "book" defines "data"
SampleMatchPairType TestMatch(const sample &book, const sample &data)
{
    // Different type samples can never match
    if(book.type != data.type)
    {
        return false;
    }
    
    return SampleMatchPairType
     (TreeCompare(book, book.alkutree, data, data.alkutree)/*,
      TreeCompare(book, book.lopputree, data, data.lopputree)*/
    );
}

void DumpSampleTree(const sample &ins, const deque<sampleitem> &tree)
{
    deque<sampleitem>::const_iterator i;
    for(i=tree.begin(); i!=tree.end(); ++i)
    {
        fprintf(stderr, "     ");
        if(ins.dutytest())
            fprintf(stderr, " Duty %d ", i->duty);
        if(ins.volutest())
        {
            fprintf(stderr, " Volume");
            if(ins.voltype!=sample::mul)fprintf(stderr, " %+d", i->voladd);
            if(ins.voltype!=sample::add)fprintf(stderr, " *%ld", i->volmul);
            fprintf(stderr, " ");
        }
        if(ins.wavetest())
        {
            fprintf(stderr, " Wavelen");
            if(ins.wavelentype!=sample::mul)fprintf(stderr, " %+d", i->wavelenadd);
            if(ins.wavelentype!=sample::add)fprintf(stderr, " *%ld", i->wavelenmul);
            fprintf(stderr, " ");
        }
        fprintf(stderr, "\n");
    }
}

void DumpSample(const sample &ins)
{
    fprintf(stderr, "   Type %s\n",
        (ins.type==SNDChannel::sqr)?"SQR"
       :(ins.type==SNDChannel::tri)?"TRI"
       :(ins.type==SNDChannel::rnd)?"RND"
       :(ins.type==SNDChannel::pcm)?"PCM":"???");
    if(ins.volutest())
        fprintf(stderr, "   Base volume: %u\n"
                        "   Volume test type: %s\n",
            ins.basevol,
            ins.voltype==sample::undefined?"UNDEFINED"
           :ins.voltype==sample::add?"ADD"
           :ins.voltype==sample::mul?"MUL":"???");
    if(ins.wavetest())
        fprintf(stderr, "   Base wavelen: %u\n"
                        "   Wavelen test type: %s\n",
            ins.basewavelen,
            ins.wavelentype==sample::undefined?"UNDEFINED"
           :ins.wavelentype==sample::add?"ADD"
           :ins.wavelentype==sample::mul?"MUL":"???");
    //fprintf(stderr, "   Length: %u samples\n", ins.tree.size());
    
    if(ins.type == SNDChannel::sqr || ins.type == SNDChannel::rnd)
    {
        struct { int len, power; } attack, decay;
        attack.len = attack.power = 0;
        for(unsigned a=1; a<ins.alkutree.size(); ++a)
        {
            const sampleitem &s = ins.alkutree[a];
            if(s.voladd <= ins.alkutree[a-1].voladd)
                break;
            attack.power = s.voladd;
            attack.len = a;
        }
        decay.len = decay.power = 0;
        for(unsigned a=1+attack.len, decaystart=a; a<ins.alkutree.size(); ++a)
        {
            const sampleitem &s1 = ins.alkutree[decaystart];
            const sampleitem &s2 = ins.alkutree[a];
            
            if(a > decaystart)
            {
                int slope = s2.voladd - ins.alkutree[a-1].voladd;
                if(slope > 0)break; // If the volume started rising again
                
                slope *= 10;
                int oldslope = decay.power*10/ decay.len;
                
                if(slope < oldslope-2 || slope > oldslope+2)
                {
                    // Dramatic change, break
                }
            }
            
            decay.len   = a-decaystart;
            decay.power = -(s2.voladd - s1.voladd);
        }
        fprintf(stderr, "  Attack(len %d,power %d) Decay(len %d,power %d)\n",
            attack.len,attack.power,
            decay.len, decay.power);
        
        switch(ins.type)
        {
            case SNDChannel::sqr:
                if(attack.power >= 3 && decay.len > 3)
                    fprintf(stderr, "    (Blow)\n");
                else if(!attack.power && decay.power)
                    fprintf(stderr, "    (Hit)\n");
                else if(!attack.power && !decay.power)
                    fprintf(stderr, "    (Flute)\n");
                else
                    fprintf(stderr, "    (String)\n");
                break;

            case SNDChannel::rnd:
                if(attack.power >= 3 && decay.len > 3)
                    fprintf(stderr, "    (Blow)\n");
                else if(!attack.power && decay.power)
                    fprintf(stderr, "    (Snare)\n");
                else
                    fprintf(stderr, "    (Snare)\n");
                break;
            default:
                break;
        }
    }
    fprintf(stderr, "  Begins:\n");
    DumpSampleTree(ins, ins.alkutree);
    /*fprintf(stderr, "  Ends:\n");
    DumpSampleTree(ins, ins.lopputree);
    */
}

unsigned AnalyzeSample(const sample &smp)
{
    /*
    fprintf(stderr, "--------------------\n");
    DumpSample(smp);
        */

    SampleMatchPairType bestmatch;
    unsigned bestmatchnum=0;
    for(unsigned a=0; a<repository.size(); ++a)
    {
        //fprintf(stderr, "COMPARING TO INS %u\n", a+1);
        SampleMatchPairType match = TestMatch(repository[a], smp);
        
        if(!match.alku.length/* || !match.loppu.length*/)
            continue;
        
        if(match.alku.length > bestmatch.alku.length
        /*
        || (match.alku.length == bestmatch.alku.length
         && match.loppu.length > bestmatch.loppu.length)
        */
          )
        {
            bestmatch    = match;
            bestmatchnum = a;
        }
    }
    
/*
    if(!bestmatch.length)
        fprintf(stderr, "->No match!\n");
    else
    {
        fprintf(stderr, "->Best match:\n"
                        "  Length %u\n", bestmatch.length);
        if(smp.wavetest())
            fprintf(stderr, "  Wavelen match: %s,%s\n",
                bestmatch.waveaddmatch?"Add":"",
                bestmatch.wavemulmatch?"Mul":"");
        if(smp.volutest())
            fprintf(stderr, "  Volume match: %s,%s\n",
                bestmatch.voladdmatch?"Add":"",
                bestmatch.volmulmatch?"Mul":"");
    }
*/
    
    if(bestmatch.alku.length)
    {
        sample &book = repository[bestmatchnum];
        sample newsample = book;
        bool changes = false;
        /* Upgrade the sample cache with possibly available matching type information */
        if(newsample.volutest() && newsample.voltype == sample::undefined)
        {
            if(bestmatch.alku.voladdmatch && !bestmatch.alku.volmulmatch)
            {
                changes = true;
                newsample.voltype = sample::add;
            }
            else if(bestmatch.alku.volmulmatch && !bestmatch.alku.voladdmatch)
            {
                changes = true;
                newsample.voltype = sample::mul;
            }
        }
        if(newsample.wavetest() && newsample.wavelentype == sample::undefined)
        {
            if(bestmatch.alku.waveaddmatch && !bestmatch.alku.wavemulmatch)
            {
                changes = true;
                newsample.wavelentype = sample::add;
            }
            else if(bestmatch.alku.wavemulmatch && !bestmatch.alku.waveaddmatch)
            {
                changes = true;
                newsample.wavelentype = sample::mul;
            }
        }
        
        if(smp.alkutree.size() > newsample.alkutree.size())
            newsample.alkutree = smp.alkutree;
        /*
        if(smp.lopputree.size() > newsample.lopputree.size())
            newsample.lopputree = smp.lopputree;
        */
        
        /*if(changes)
        {
            repository.push_back(newsample);
            return repository.size();
        }
        else*/
        {
            //fprintf(stderr, "UPDATED INSTRUMENT %u:\n", bestmatchnum+1);
            //DumpSample(newsample);
            book = newsample;
        }
        return bestmatchnum+1;
    }
    else
    {
        sample tmp = smp;
        repository.push_back(tmp);
        
        //fprintf(stderr, "CACHED INSTRUMENT %u:\n", repository.size());
        //DumpSample(tmp);
        
        /* None was found */
        return repository.size();
    }
}
